#include "qt-demo.h"
#include "ui_qt-demo.h"

#include <QApplication>
#include <QFileDialog>

#include <QFile>
#include <QXmlStreamReader>
#include <QXmlStreamAttribute>

#include <QClipboard>
#include <QDrag>
#include <QMimeData>
#include <QIcon>
#include <QGraphicsSceneDragDropEvent>

#include <QScriptEngine>
#include <QScriptEngineDebugger>

#include <cassert>

MainWindow * global_win = NULL;

/* parameter list */

int conv (QString s, int default_value)
{
    bool ok;
    int n = s.toInt (&ok);
    if (!ok)
        n = default_value;
    return n;
}

int ParamList::number (QString key, int default_value)
{
    return conv (this->value (key), default_value);
}

/* Shape */

const QString shape_format = "application/x-shape";

void shapeDragEnter (QGraphicsSceneDragDropEvent *event)
{
    const QMimeData * data = event->mimeData();
    if (data->hasColor() || data->hasFormat (shape_format))
        event->setAccepted (true);
}

void shapeDrop (QAbstractGraphicsShapeItem * shape, QGraphicsSceneDragDropEvent *event)
{
   const QMimeData * data = event->mimeData();
   if (data->hasColor())
   {
       QColor c =  data->colorData().value <QColor> ();
       if (event->proposedAction() == Qt::MoveAction) /* shift */
          shape->setPen (c);
       else
          shape->setBrush (c);
   }
   if (data->hasFormat (shape_format))
   {
       QString code = data->data (shape_format);
       if (global_win != NULL)
          global_win->addGraphicsItem (shape, code, event->proposedAction() == Qt::MoveAction);
   }
}

/* Rectangle */

void Rectangle::dragEnterEvent (QGraphicsSceneDragDropEvent *event)
{
    shapeDragEnter (event);
}

void Rectangle::dropEvent (QGraphicsSceneDragDropEvent *event)
{
    shapeDrop (this, event);
}

Rectangle::Rectangle()
{
    setAcceptDrops (true);
}

/* Ellipse */

void Ellipse::dragEnterEvent (QGraphicsSceneDragDropEvent *event)
{
    shapeDragEnter (event);
}

void Ellipse::dropEvent (QGraphicsSceneDragDropEvent *event)
{
    shapeDrop (this, event);
}

Ellipse::Ellipse ()
{
    setAcceptDrops (true);
}

/* Color Button */

ColorButton::ColorButton (QString s)
  : name (s), color (QColor (s))
{
    setText (s);
    setToolTip (s);
    setIcon (createPixmap ());
}

void ColorButton::mousePressEvent(QMouseEvent *event)
{
    QDrag * drag = new QDrag (this);
    QMimeData * data = new QMimeData ();
    data->setText (name);
    data->setColorData (color);
    drag->setMimeData(data);
    drag->setPixmap (createPixmap (true));
    drag->start(Qt::CopyAction | Qt::MoveAction);
}

QPixmap ColorButton::createPixmap (bool white)
{
    QPixmap pixmap (24, 24);
    if (white)
       pixmap.fill (QColor (255, 255, 255, 255));
    else
       pixmap.fill (QColor (0, 0, 0, 0));

    QPainter painte (& pixmap);
    painte.setPen (color);
    painte.setBrush (color);
    painte.drawEllipse (2, 2, 20, 20);

    return pixmap;
};

/* Shape Button */

ShapeButton::ShapeButton (QString n, QString c)
  : name (n),
    code (c)
{
    code.replace ('\'', '\"');
    code = "<data>" + code + "</data>"; // add top level element
    setText (name);
    setToolTip (code);
}

void ShapeButton::mousePressEvent (QMouseEvent *event)
{
    QDrag * drag = new QDrag (this);
    QMimeData * data = new QMimeData ();
    data->setData (shape_format, code.toAscii());
    drag->setMimeData (data);
    drag->start ();
}


/* properties */

const int name_inx = 1;
const int type_inx = 2;
const int pen_inx = 3;
const int brush_inx = 4;

QGraphicsItem * createElement (QString type, ParamList params)
{
    QAbstractGraphicsShapeItem * shape = NULL;
    if (type == "rectangle")
        shape = new Rectangle;
    else if (type == "ellipse")
        shape = new Ellipse;

    if (shape != NULL)
    {
        QString s = params ["name"];
        shape->setData (name_inx, s);

        shape->setData (type_inx, type);

        int x = params.number ("x", 0);
        int y = params.number ("y", 0);
        shape->setPos (x, y);

        int w = params.number ("width", 20);
        int h = params.number ("height", 20);
        QGraphicsRectItem * r = dynamic_cast <QGraphicsRectItem *> (shape);
        if (r != NULL)
            r->setRect (0, 0, w, h);
        QGraphicsEllipseItem * e = dynamic_cast <QGraphicsEllipseItem *> (shape);
        if (e != NULL)
            e->setRect (0, 0, w, h);

        s = params ["pen"];
        shape->setData (pen_inx, s);
        if (s != "")
           shape->setPen (QColor (s));

        s = params ["brush"];
        shape->setData (brush_inx, s);
        if (s != "")
           shape->setBrush (QColor (s));
     }

    return shape;
}

void updateAttribute (QGraphicsItem * elem, QString attr, QString value)
{
    assert (elem != NULL);

    QAbstractGraphicsShapeItem * shape = dynamic_cast <QAbstractGraphicsShapeItem *> (elem);
    QGraphicsRectItem * r = dynamic_cast <QGraphicsRectItem *> (elem);
    QGraphicsEllipseItem * e = dynamic_cast <QGraphicsEllipseItem *> (elem);

     if (attr == "name")
        elem->setData (name_inx, value);

     if (attr == "x")
        shape->setX (conv (value, 0));

     if (attr == "y")
        shape->setY (conv (value, 0));

     if (attr == "width" || attr == "height")
     {
        QRectF size;
        if (r != NULL) size = r->rect ();
        if (e != NULL) size = e->rect ();

        int v = conv (value, 0);
        if (attr == "width")
           size.setWidth (v);
        else
           size.setHeight (v);

        if (r != NULL) r->setRect (size);
        if (e != NULL) e->setRect (size);
     }

     if (shape != NULL)
     {
        if (attr == "pen")
        {
           shape->setPen (QColor (value));
           shape->setData (pen_inx, value);
        }

         if (attr == "brush")
         {
            shape->setBrush (QColor (value));
            shape->setData (pen_inx, value);
         }
     }
}

QString num_to_str (int n)
{
   return QString::number (n);
}

ParamList recallAttributes (QGraphicsItem * elem)
{
   ParamList param;

   assert (elem != NULL);
   QAbstractGraphicsShapeItem * shape = dynamic_cast <QAbstractGraphicsShapeItem *> (elem);
   QGraphicsRectItem * r = dynamic_cast <QGraphicsRectItem *> (elem);
   QGraphicsEllipseItem * e = dynamic_cast <QGraphicsEllipseItem *> (elem);

   param ["name"] = elem->data (name_inx).toString ();

   if (shape != NULL)
   {
      param ["x"] = num_to_str (shape->x());
      param ["y"] = num_to_str (shape->y());
   }

   QRectF size;
   if (r != NULL) size = r->rect ();
   if (e != NULL) size = e->rect ();
   if (r != NULL || e != NULL)
   {
      param ["width"] = num_to_str (size.width ());
      param ["height"] = num_to_str (size.height ());
   }

   param ["pen"] = elem->data (pen_inx).toString ();
   param ["brush"] = elem->data (brush_inx).toString ();

   return param;
}

/* tree and scene */

TreeNode * createTreeNode (QGraphicsItem * item)
{
   TreeNode * node = new TreeNode;

   node->elem = item;

   QString s = item->data (type_inx).toString();
   QString n = item->data (name_inx).toString();
   if (n != "")
      s = s + " " + n;
   node->setText (0, s);

   node->setForeground (0, QColor ("blue"));

   return node;
}

QGraphicsItem * getElement (QTreeWidgetItem * item)
{
    QGraphicsItem * result = NULL;

    TreeNode * node = dynamic_cast <TreeNode *> (item);
    if (node != NULL)
        result = node->elem;

    return result;
}

TreeNode * findNode (QTreeWidget * tree, QGraphicsItem * item)
{
   TreeNode * result = NULL;
   if (item != NULL)
   {
      QTreeWidgetItemIterator it (tree);
      while (*it && result == NULL)
      {
          TreeNode * node = dynamic_cast <TreeNode *> (*it);
          if (node != NULL && node->elem == item)
              result = node;
          ++it;
      }
   }
   return result;
}

void addIntoScene (QGraphicsItem * item, QTreeWidgetItem * branch, QGraphicsScene * scene)
{
    if (item != NULL)
    {
        item->setFlag (QGraphicsItem::ItemIsMovable);
        item->setFlag (QGraphicsItem::ItemIsSelectable);

        QGraphicsItem * target = getElement (branch);

        if (target == NULL)
            scene->addItem (item);
        else
            item->setParentItem (target);
    }
}

/* input / output */

void readXml (QXmlStreamReader & reader, QTreeWidgetItem * branch, QGraphicsScene * scene)
{
    int level = 0;
    QTreeWidgetItem * start_branch = branch;

    while (! reader.atEnd())
    {
        if (reader.isStartElement())
        {
            if (level > 0) // ignore top level element
            {
                QString type = reader.name().toString();

                ParamList params;
                // for (QXmlStreamAttribute a : reader.attributes())
                for (auto a : reader.attributes())
                {
                   QString attr = a.name().toString();
                   QString value = a.value().toString();
                   params [attr] = value;
                }

                QGraphicsItem * elem = createElement (type, params);
                assert (elem != NULL);

                TreeNode * node = createTreeNode (elem);
                assert (node != NULL);

                // add into tree
                branch->addChild (node);

                addIntoScene (elem, branch, scene);

                branch = node; // new branch
             }
             level ++;
        }
        else if (reader.isEndElement())
        {
            level --;
            if (level > 0)
               branch = branch->parent();
            if (branch == NULL)
               branch = start_branch;
        }
        reader.readNext();
     }
}

void MainWindow::writeItem (QXmlStreamWriter & writer, QTreeWidgetItem * item)
{
    QGraphicsItem * elem = getElement (item);
    if (elem != NULL)
    {
        QString type = elem->data (type_inx).toString ();
        writer.writeStartElement (type);

        ParamList param = recallAttributes (elem);
        QMapIterator <QString, QString> iter (param);
        while (iter.hasNext())
        {
            iter.next();
            QString attr = iter.key ();
            QString value = iter.value ();
            writer.writeAttribute (attr, value);
        }

        int cnt = item->childCount ();
        for (int inx = 0; inx < cnt; inx ++)
        {
            writeItem (writer, item->child (inx));
        }

        writer.writeEndElement ();
    }
}

void MainWindow::readFile (QString fileName)
{
    QFile file (fileName);
    if (file.open (QFile::ReadOnly | QFile::Text ))
    {
        QXmlStreamReader reader (&file);
        QTreeWidgetItem * branch = ui->treeWidget->invisibleRootItem();
        readXml (reader, branch, scene);
        ui->treeWidget->expandAll ();
    }
}

void MainWindow::writeFile (QString fileName)
{
    QFile file (fileName);
    if (file.open (QFile::WriteOnly | QFile::Text))
    {
        QXmlStreamWriter writer (&file);
        writer.setAutoFormatting (true);
        writer.writeStartDocument ();
        writer.writeStartElement ("data"); // top level element

        int cnt = ui->treeWidget->topLevelItemCount ();
        for (int inx = 0; inx < cnt; inx ++)
        {
            QTreeWidgetItem * item = ui->treeWidget->topLevelItem (inx);
            writeItem (writer, item);
        }

        writer.writeEndElement ();
        writer.writeEndDocument ();
    }
}

void MainWindow::on_actionOpen_triggered()
{
    QString s = QFileDialog::getOpenFileName (this, "Open XML file");
    if (s != "")
       readFile (s);
}

void MainWindow::on_actionSave_triggered()
{
    QString s = QFileDialog::getSaveFileName (this, "Save XML file");
    if (s != "")
       writeFile (s);
}

/* drag and drop */

void MainWindow::addGraphicsItem (QGraphicsItem * shape, QString code, bool plus, int relx, int rely)
{
   QTreeWidgetItem * branch = findNode (ui->treeWidget, shape);
   if (plus && branch != NULL)
       branch = branch->parent ();
   if (branch == NULL)
       branch = ui->treeWidget->invisibleRootItem();
   QXmlStreamReader reader (code);
   readXml (reader, branch, scene);
}

/* clipboard */

void MainWindow::on_actionCopy_triggered()
{
   QString code = "";
   QXmlStreamWriter writer (& code);

   writer.setAutoFormatting (true);
   writer.writeStartDocument ();
   writer.writeStartElement ("data"); // top level element

   for (QGraphicsItem * item : scene->selectedItems ())
   {
       TreeNode * node = findNode (ui->treeWidget, item);
       if (node != NULL)
          writeItem (writer, node);
   }

   writer.writeEndElement ();
   writer.writeEndDocument ();

   QMimeData * data = new QMimeData;
   data->setData (shape_format, code.toAscii());

   QClipboard * clip = QApplication::clipboard ();
   clip->setMimeData (data);
}

void MainWindow::on_actionPaste_triggered()
{
   QClipboard * clip = QApplication::clipboard ();
   const QMimeData * data = clip->mimeData ();
   if (data->hasFormat (shape_format))
   {
      QString code = data->data (shape_format);
      addGraphicsItem (NULL, code, true);
   }


}
/* tree and property table */

void MainWindow::on_treeWidget_itemClicked (QTreeWidgetItem *item, int column)
{
    QGraphicsItem * elem = getElement (item);
    if (elem != NULL)
    {
        ui->statusBar->showMessage ("click");
        showTable (elem);
    }
}

void MainWindow::showTable (QGraphicsItem * elem)
{
    current_element = NULL; // disable on_tableWidget_itemChanged

    ui->tableWidget->clear ();
    ui->tableWidget->setColumnCount (2);
    ui->tableWidget->setHorizontalHeaderLabels (QStringList () << "Name" << "Value");
    ui->tableWidget->verticalHeader()->hide();

    int cnt = 0;
    ParamList param = recallAttributes (elem);
    QMapIterator <QString, QString> iter (param);
    while (iter.hasNext())
    {
        iter.next();
        QString attr = iter.key ();
        QString value = iter.value ();

        ui->tableWidget->setRowCount(cnt+1);

        QTableWidgetItem * field = new QTableWidgetItem;
        field->setText (attr);
        ui->tableWidget->setItem (cnt, 0, field);

        field = new QTableWidgetItem;
        field->setText (value);
        ui->tableWidget->setItem (cnt, 1, field);

        field->setData (Qt::UserRole, attr);
        if (attr == "pen" or attr == "brush")
           field->setData (Qt::DecorationRole, QColor (value));

        cnt ++;
    }

    current_element = elem;
}

void MainWindow::on_tableWidget_itemChanged (QTableWidgetItem * item)
{
    if (current_element != NULL)
    {
       QString attr = item->data (Qt::UserRole).toString();
       QString value = item->text ();
       ui->statusBar->showMessage ("update " + attr + " = " + value);
       updateAttribute (current_element, attr, value);
    }
}

/* sctipt */

void MainWindow::on_actionRun_triggered()
{
    QString source = ui->input->toPlainText ();
    QScriptEngine engine;

    QScriptEngineDebugger debugger;
    debugger.attachTo (& engine);

    engine.globalObject().setProperty ("output", engine.newQObject (ui->output));

    for (QGraphicsItem * item : scene->items ())
    {
       QObject * obj = dynamic_cast <QObject *> (item);
       if (obj != NULL)
       {
          QString name = item->data (name_inx).toString();
          if (name != "")
          {
             engine.globalObject().setProperty (name, engine.newQObject (obj));
             ui->statusBar->showMessage ("obj " + name);
          }
       }
    }

    engine.evaluate (source);
}

/* MainWindow */

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    scene (NULL),
    current_element (NULL)
{
    ui->setupUi(this);

    global_win = this;

    ui->statusBar->showMessage(QString ("Qt ") + qVersion());

    /* setup tools */

    QString code = "<rectangle x='20' y='50' width='60' height='20' pen='red' brush='orange' />";
    ui->mainToolBar->addWidget (new ShapeButton ("rectangle", code));

    code = "<ellipse x='20' y='20' width='20' height='20' pen='red' brush='blue' />";
    ui->mainToolBar->addWidget (new ShapeButton ("ellipse", code));

    code = "<rectangle x='10' y='10' width='100' height='100' pen='red' brush='yellow'>"
           "  <rectangle x='20' y='50' width='60' height='20' pen='red' brush='orange'/>"
            "</rectangle>" ;
    ui->mainToolBar->addWidget (new ShapeButton ("block", code));

    ui->mainToolBar->addSeparator ();

    /* setup colors */

    QList <QString> names;
    names << "red" << "blue" << "green" << "yellow" << "orange";

    for (auto name : names)
    {
       ColorButton * b = new ColorButton (name);
       ui->mainToolBar->addWidget (b);
    }

    /* setup scene */

    scene = new QGraphicsScene;
    ui->graphicsView->setScene (scene);

    scene->setSceneRect (0, 0, 1280, 1024);

    readFile (":/data.xml");

    /* script */

    QFile file (":/input.js");
    if (file.open (QFile::ReadOnly | QFile::Text))
       ui->input->setText (file.readAll());
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_actionQuit_triggered()
{
   this->close ();
}

int main(int argc, char *argv[])
{
    QApplication a(argc, argv);
    MainWindow w;
    w.show();
    return a.exec();
}





